/*

Copyright (c) 2003-2004, AXE, Inc.  All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "libhttp.h"

#define BUFPAGESIZE 1024
#define HTTP_VERSION "HTTP/1.1"

#define putcToBuf(conn, c) \
(((conn)->bufIndex < (conn)->bufSize) ? (conn)->lineBuf[(conn)->bufIndex++]\
 = (c) : extendPutcToBuf((conn), (c)))

/*
    Initialize
*/
int HTTPInit(int fd, Conn *conn)
{
    if ((conn->rfp = fdopen(fd,"rb")) == NULL) {
	return -1;
    }
    if ((conn->wfp = fdopen(fd,"wb")) == NULL) {
	fclose(conn->rfp);
	return -1;
    }
    /* initialize buffer */
    if ((conn->lineBuf = malloc(BUFPAGESIZE)) == NULL) {
	fclose(conn->rfp);
	fclose(conn->wfp);
	return -1;
    }
    conn->bufSize = BUFPAGESIZE;
    conn->bufIndex = 0;

    conn->method = NULL;
    conn->uri = NULL;
    conn->version = NULL;
    conn->headers = NULL;
    conn->length = 0;
    conn->body = NULL;
    return 0;
}

/*
    find specified header and return it's value
*/
char *HTTPFindHeader(Conn *conn, char *name)
{
    Header *hp;

    for (hp = conn->headers;hp != NULL;hp = hp->next) {
	if (strcmp(hp->name,name) == 0) return hp->value;
    }
    return NULL;
}

/*
    allocate new header structure
*/
static Header *newHeader(void)
{
    Header *nhp;

    if ((nhp = malloc(sizeof(Header))) == NULL) {
	return NULL;
    }
    nhp->next = NULL;
    nhp->name = NULL;
    nhp->value = NULL;
    return nhp;
}

/*
    free header structure
*/
static void freeHeader(Header *hp)
{
    if (hp->name != NULL) free(hp->name);
    if (hp->value != NULL) free(hp->value);
    free(hp);
}

/*
    free header structure list
*/
static void freeHeaders(Header *hp)
{
    while (hp != NULL) {
	Header *nhp = hp->next;

	freeHeader(hp);
	hp = nhp;
    }
}

/*
    clean connection structure
    Note: structure itself is not freed
*/
void HTTPClean(Conn *conn)
{
    conn->bufIndex = 0;
    if (conn->method != NULL) {
	free(conn->method);
	conn->method = NULL;
    }
    if (conn->uri != NULL) {
	free(conn->uri);
	conn->uri = NULL;
    }
    if (conn->version != NULL) {
	free(conn->version);
	conn->version = NULL;
    }
    if (conn->headers != NULL) {
	freeHeaders(conn->headers);
	conn->headers = NULL;
    }
    if (conn->body != NULL) {
	free(conn->body);
	conn->body = NULL;
    }
}

/*
    free all member of Connection structure
    Note: close fd
*/
void HTTPDestroy(Conn *conn)
{
    HTTPClean(conn);
    if (conn->lineBuf != NULL) {
	free(conn->lineBuf);
	conn->lineBuf = NULL;
    }
    fclose(conn->rfp);
    fclose(conn->wfp);
}

/*
    insert header structure into the list
*/
static void insHeader(Conn *conn, Header *hp)
{
    hp->next = conn->headers;
    conn->headers = hp;
}

/*
    put a character into the buffer
    if the buffer is full, it is extended.
*/
static int extendPutcToBuf(Conn *conn, int c)
{
    if ((conn->lineBuf = realloc(conn->lineBuf, conn->bufSize+BUFPAGESIZE))
       == NULL) {
	return -1;
    }
    conn->bufSize += BUFPAGESIZE;
    conn->lineBuf[conn->bufIndex++] = c;
    return c;
}

/*
    wtite HTTP response
*/
int HTTPWriteResponse(Conn *conn, int code, char *reason,
  int nHeader, char **name, char **value,
  int bodyLen, char *body)
{
    int i;

    fprintf(conn->wfp,HTTP_VERSION " %d %s\r\n",code,reason);
    for (i = 0;i < nHeader;i++) {
	fprintf(conn->wfp,"%s: %s\r\n",name[i],value[i]);
    }
    if (bodyLen > 0) {
	fprintf(conn->wfp,"Content-Length: %d\r\n",bodyLen);
	fprintf(conn->wfp,"\r\n");
	fwrite(body,1,bodyLen,conn->wfp);
    } else {
	fprintf(conn->wfp,"Content-Length: 0\r\n");
	fprintf(conn->wfp,"\r\n");
    }
    fflush(conn->wfp);
    return 0;
}

/*
    skip space characters
*/
static int skipSP(Conn *conn)
{
    int c;

    while ((c = fgetc(conn->rfp)) >= 0 && (c == ' ' || c == '\t'));
    ungetc(c,conn->rfp);
}

/*
    get a word into the buffer
*/
static int getWord(Conn *conn)
{
    int c;

    while ((c = fgetc(conn->rfp)) >= 0 && c != ' ' && c != '\t'
          && c != '\r' && c != '\n') {
	if (putcToBuf(conn,c) < 0) {
	    return -1;
	}
    }
    ungetc(c,conn->rfp);
    if (putcToBuf(conn,'\0') < 0) {
	return -1;
    }
    return 0;
}

/*
    copy the data in the buffer of connection structure
      to the specified buffer
*/
static int copyBuf(Conn *conn, char **sp)
{
    if ((*sp = strdup(conn->lineBuf)) == NULL) {
	return -1;
    }
    conn->bufIndex = 0;
    return 0;
}

/*
    send "Bad request" response
*/
static int sendBadRequest(Conn *conn)
{
    return HTTPWriteResponse(conn,400,"Bad Request",0,NULL,NULL,0,NULL);
}

/*
    send "Length Required" response
*/
static int sendLengthRequired(Conn *conn)
{
    return HTTPWriteResponse(conn,411,"Length Required",0,NULL,NULL,0,NULL);
}

/*
    skip characters until CRLF
*/
static int skipToCRLF(Conn *conn)
{
    int c;

    while ((c = fgetc(conn->rfp)) >= 0 && c != '\n' && c != '\r');
    if (c == '\r') {
	c = fgetc(conn->rfp);
	if (c != '\n' && c != '\r') {
	    ungetc(c,conn->rfp);
	}
    }
    return 0;
}

/*
    get HTTP method string
*/
static int getMethod(Conn *conn)
{
    if (getWord(conn) < 0) return -1;
    if (copyBuf(conn,&conn->method) < 0) {
	return -1;
    }
    return 0;
}

/*
    get request line URI string
*/
static int getURI(Conn *conn)
{
    if (getWord(conn) < 0) return -1;
    if (copyBuf(conn,&conn->uri) < 0) {
	return -1;
    }
    return 0;
}

/*
    get HTTP Version string
*/
static int getVersion(Conn *conn)
{
    if (getWord(conn) < 0) return -1;
    if (copyBuf(conn,&conn->version) < 0) {
	return -1;
    }
    return 0;
}

/*
    get Request Line
*/
static int getRequestLine(Conn *conn)
{
    if (getMethod(conn) < 0) {
	return -1;
    }
    if (skipSP(conn) < 0) {
	return -1;
    }
    if (getURI(conn) < 0) {
	return -1;
    }
    if (skipSP(conn) < 0) {
	return -1;
    }
    if (getVersion(conn) < 0) {
	return -1;
    }
    if (skipToCRLF(conn) < 0) {
	return -1;
    }
    return 0;
}

/*
    get field name
*/
static int getFieldName(Conn *conn)
{
    int c;

    if ((c = fgetc(conn->rfp)) == '\r' || c == '\n') {
	/* end of headers */
	if (c == '\r') {
	    c = fgetc(conn->rfp);
	    if (c != '\n' && c != '\r') {
		ungetc(c,conn->rfp);
	    }
	}
	return 1;
    }
    if (c < 0) {
	/* found eof before enf of headers  */
	sendBadRequest(conn);
	return -1;
    }
    ungetc(c,conn->rfp);
    while ((c = fgetc(conn->rfp)) >= 0 && c != ' ' && c != '\t'
          && c != '\r' && c != '\n' && c != ':') {
	if (islower(c)) {
	    c = toupper(c);
	}
	if (putcToBuf(conn,c) < 0) {
	    return -1;
	}
    }
    if (c != ':') {
	sendBadRequest(conn);
	return -1;
    }
    if (putcToBuf(conn,'\0') < 0) {
	return -1;
    }
    return 0;
}

/*
 get field value
*/
static int getFieldValue(Conn *conn)
{
    int c;
    int quoted = 0;

    if (skipSP(conn) < 0) {
	return -1;
    }
    do {
	while ((c = fgetc(conn->rfp)) >= 0
	  && (quoted  || (c != '\r' && c != '\n'))) {
	    if (c == '"') {
		quoted = !quoted;
	    } else if (islower(c)) {
		c = toupper(c);
	    }
	    if (putcToBuf(conn,c) < 0) {
		return -1;
	    }
	}
    /* check continuation line */
    } while ((c = fgetc(conn->rfp)) >= 0 && (c == ' ' || c == '\t'));
    ungetc(c,conn->rfp);
    if (putcToBuf(conn,'\0') < 0) {
	return -1;
    }
    return 0;
}

/*
    get HTTP header
*/
static int getHeader(Conn *conn)
{
    int r;
    Header *nhp;

    if ((r = getFieldName(conn)) < 0) {
	return -1;
    }
    if (r != 0) return r;

    if ((nhp = newHeader()) == NULL) {
	return -1;
    }
    if (copyBuf(conn,&nhp->name) < 0) {
	freeHeader(nhp);
	return -1;
    }

    if (getFieldValue(conn) < 0) {
	return -1;
    }
    if (copyBuf(conn,&nhp->value) < 0) {
	freeHeader(nhp);
	return -1;
    }
    insHeader(conn,nhp);
    if (skipToCRLF(conn) < 0) {
	return -1;
    }
    return 0;
}

/*
    get HTTP BODY
*/
static int getBody(Conn *conn)
{
    char *slen;
    int len = 0;
    int i;

    if ((slen = HTTPFindHeader(conn,"CONTENT-LENGTH")) != NULL) {
	conn->length = len = atoi(slen);
    }
    if ((conn->body = malloc(len)) == NULL) {
	return -1;
    }
    for (i = 0;i < len;i++) {
	int c;

	if ((c = fgetc(conn->rfp)) < 0) {
	    sendBadRequest(conn);
	    return -1;
	}
	conn->body[i] = c;
    }
    return 0;
}

/*
    get a HTTP request
*/
int HTTPGetReq(Conn *conn)
{
    int r;

    HTTPClean(conn);
    /* get request line */
    if (getRequestLine(conn) < 0) {
	return -1;
    }
    /* get headers */
    while ((r = getHeader(conn)) == 0);
    if (r < 0) { /* error */
	return -1;
    }
    /* get body */
    if (getBody(conn) < 0) {
	return -1;
    }
    return 0;
}

/*
    parse option in a request URI
    opt: pointer to the start of the option

    return option's name
	when end of string return NULL:w
    *value: option's value
    *next : pointer to the start of the next option
*/
char *HTTPParseOpt(char *opt, char **value, char **next)
{
    char *p;
    char *r;
    char *name;

    if (opt == NULL) return NULL;
    if ((r = strchr(opt,'&')) != NULL) {
	*r++ = '\0';
    }
    name = opt;
    if ((p = strchr(opt,'=')) == NULL) {
	p = p+strlen(opt);
    } else {
	*p++ = '\0';
    }
    *value = p;
    *next = r;
    return name[0] == '\0' ? NULL : name;
}
