/*

Copyright (c) 2003-2004, AXE, Inc.  All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
/*
 * oppd-listen.c
 * open-printing printer daemon listener
 */
#include <stdio.h>
#include <fcntl.h>
#include <memory.h>
#include <malloc.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netdb.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <errno.h>

#ifndef WITHOUT_BIDI
#include "bidiPlugin.h"
#define BIDI_JOB_ID 1
#endif

#ifndef DELJOB_ON_ERR
#define DELJOB_ON_ERR 0
#endif

#include "oppd_common.h"
#include "opvp_rpc_core.h"
#include "opvp_rpc_server.h"
#include "oppd_conf.h"

#define OPPD_OUTPUT_TIMEOUT 10

/*
 *
 */
oppd_signal	sig[] = {
	{SIGCHLD, NULL},
	{SIGTERM, NULL},
	{SIGPIPE, NULL},
	{SIGALRM, NULL},
	{-1     , NULL}
};
int		force_abort = 0;
char		printer[OPPD_BUF_LEN];
char		device[OPPD_BUF_LEN];
char		bidi[OPPD_BUF_LEN];

/*
 * signal handler
 */
static
void		sig_handler(
	int		signo
#ifdef USE_SIGINFO
	,
	siginfo_t	*siginfo,
	void		*ucntextptr
#endif
	)
{
	pid_t		pid;
	int		stat;

	switch (signo) {
		case	SIGCHLD :
			while((pid = waitpid(-1, &stat, WNOHANG)) > 0);
			break;
		case	SIGALRM :
			/* ignore */
			break;
		case	SIGTERM :
			force_abort = 1;
			break;
		case	SIGPIPE :
			/* ignore */
			break;
		default :
			/* ignore */
			break;
	}
	return;
}

/*
 * initialize signals
 */
static
int		sig_init(
	void
	)
{
	int		ecode = 0;
	int		i = 0;
struct	sigaction	act;
struct	sigaction	old;

	while (sig[i].signo != -1) {
		act.sa_flags = SA_RESTART;
#ifdef USE_SIGINFO
		act.sa_sigaction = sig_handler;
		act.sa_flags |= SA_SIGINFO;
#else
		act.sa_handler = sig_handler;
#endif
		sigemptyset(&act.sa_mask);
		if (sigaction(sig[i].signo, &act, &old) < 0) {
			ecode = -1;
			break;
		}
		sig[i].act = malloc(sizeof(struct sigaction));
		if (!sig[i].act) {
			ecode = -1;
			break;
		}
		memcpy(sig[i].act, &old, sizeof(struct sigaction));
		i++;
	}

	return ecode;
}

/*
 * finalize signals
 */
static
void		sig_fin(
	void
	)
{
	int		i = 0;

	while (sig[i].signo != -1) {
		if (!sig[i].act) {
			sigaction(sig[i].signo, sig[i].act, NULL);
			free(sig[i].act);
			sig[i].act = NULL;
		}
		i++;
	}

	return;
}

/*
 * open listener sockets
 */
static
int		open_sock(
	int		*sock
	)
{
	int		ecode = 0;
struct	sockaddr_un	addr;

	*sock = socket(PF_UNIX, SOCK_STREAM, 0);
	if (*sock != -1) {
		bzero(&addr, sizeof(struct sockaddr_un));
		addr.sun_family = PF_UNIX;
		strcpy(addr.sun_path, "/tmp/OPPD-");
		strcat(addr.sun_path, printer);

		unlink(addr.sun_path);

		bind(*sock, (struct sockaddr *)&addr, sizeof(addr));
	} else {
		ecode = -1;
	}

	return ecode;
}

/*
 * close listener sockets
 */
static
void		close_sock(
	int		*sock
	)
{
	char		path[OPPD_BUF_LEN];

	if (*sock != -1) {
		close(*sock);
		*sock = -1;
	}

	/* unlink socket-file */
	strcpy(path, "/tmp/OPPD-");
	strcat(path, printer);
	unlink(path);

	return;
}

static int		dev_fd = -1;
static int bidiCanWrite = 0; /* support bidiWrite */
static int bidiCanJob = 0; /* support bidiStartJob, bidiEndJob */
static int bidiCanCtrl = 0; /* support bidiCtrl */
static int bidiRfd = -1;
static int bidiWfd = -1;
static BidiC *bidiC = NULL; /* bidi Context handle */
static int bidiJob = 0; /* start job counter */
static int eprogress = 0; /* bidi is in the EPROGRESS state */
static int bidiInWriting = 0; /* bidi is in a writing state */

/* output process states */

#define PDST_NONE       0x0 /* no request */
#define PDST_PRT        0x1 /* printing request */
#define PDST_ST         0x2 /* processing status request */
#define PDST_NOPRT_NOST (PDST_NONE)        /* no printing job,
                                              no status request */
#define PDST_PRT_NOST   (PDST_PRT)         /* printing, no status request */
#define PDST_NOPRT_ST   (PDST_ST)          /* no printing job,
                                              processing status request */
#define PDST_PRT_ST     (PDST_PRT|PDST_ST) /* printing,
                                              processing status request */

static int oppdState = PDST_NOPRT_NOST;

static	int		spl_fd = -1; /* spooled file descriptor */
static	int		spl_size = -1; /* spooled file size */
static	int		spl_len = -1; /* spooled file size */
static	int		read_len; /* read spooled file data size */
static	int		want_write = 0; /* I want to write data in the buffer
                                           to the printer */
static	void		*rpcHandle = NULL; /* RPC Handle */
static	int		write_fd = -1; /* file descriptor for printing */
static	oppd_job	*job = NULL; /* printing job handle */
static	int		sock_fd = -1; /* socket descriptor */

#define	DATA_BUF_LEN	(OPPD_BUF_LEN*4)
static	char		printingBuf[DATA_BUF_LEN]; /* buffer for printing */

/*
    open device file
*/
static int openDevice(void)
{
    if (dev_fd >= 0) return 0; /* return if already open */
    if ((dev_fd = open(device,O_RDWR /*|O_NONBLOCK*/)) < 0) {
	return -1;
    }
    if (flock(dev_fd, LOCK_EX) < 0) {
	return -1;
    }
    return 0;
}

/*
    close device file
*/
static void closeDevice(void)
{
    if (dev_fd >= 0) {
	flock(dev_fd, LOCK_UN);
	close(dev_fd);
	dev_fd = -1;
    }
}

/*
    initialize BIDI object
*/
static int initBidi(void)
{
    int r;

    if (openDevice() < 0) return -1;
#ifndef WITHOUT_BIDI
    if (bidiC == NULL) {
	alarm(OPPD_OUTPUT_TIMEOUT);
	bidiC = bidiNew(bidi,dev_fd,dev_fd,NULL);
	alarm(0);
	if (bidiC == NULL) {
	    goto close_dev;
	}
	if ((r = bidiGetCap(bidiC,BIDI_CAP_WRITE)) == BIDI_ERROR) {
	    goto dest_bidi;
	}
	if (r == BIDI_TRUE) {
	    bidiCanWrite = 1;
	} else {
	    bidiCanWrite = 0;
	}
	if ((r = bidiGetCap(bidiC,BIDI_CAP_JOB)) == BIDI_ERROR) {
	    goto dest_bidi;
	}
	if (r == BIDI_TRUE) {
	    bidiCanJob = 1;
	} else {
	    bidiCanJob = 0;
	}
	if ((r = bidiGetCap(bidiC,BIDI_CAP_CTRL)) == BIDI_ERROR) {
	    goto dest_bidi;
	}
	if (r == BIDI_TRUE) {
	    bidiCanCtrl = 1;
	} else {
	    bidiCanCtrl = 0;
	}
	if ((bidiRfd = bidiGetReadFD(bidiC)) == BIDI_ERROR) {
	    goto dest_bidi;
	}
	if ((bidiWfd = bidiGetWriteFD(bidiC)) == BIDI_ERROR) {
	    goto dest_bidi;
	}
	if (bidiCanWrite) {
	    write_fd = bidiWfd;
	} else {
	    write_fd = dev_fd;
	}
    }
#else
    write_fd = dev_fd;
#endif
    return 0;
dest_bidi:
    alarm(OPPD_OUTPUT_TIMEOUT);
    bidiDestroy(bidiC);
    alarm(0);
    bidiC = NULL;
close_dev:
    closeDevice();
    return -1;
}

/*
 * initialize
 */
static
int		init(
	int		*sock
	)
{
	int		ecode = 0;
	int r;

	/* initialize signal */
	if (!ecode) ecode = sig_init();

 	/* open listener sockets */
	if (!ecode) ecode = open_sock(sock);

	if (ecode) return ecode;
	return ecode;
}

/*
    finalize BIDI object
*/
static void finBidi(void)
{
#ifdef DEBUG
    fprintf(stderr,"finBidi\n");
#endif
#ifndef WITHOUT_BIDI
    if (bidiC != NULL) {
	alarm(OPPD_OUTPUT_TIMEOUT);
	bidiDestroy(bidiC);
	alarm(0);
	bidiC = NULL;
    }
    /* clear bidi state variables */
    bidiJob = 0;
    eprogress = 0;
#endif
    closeDevice();
}

/*
    process when BIDI_ERROR
*/
static void procBidiError(void)
{
    finBidi();
    /* force reset state */
    oppdState = PDST_NOPRT_NOST;
}

/*
 * finalize
 */
static
void		fin(
	int		*sock
	)
{

 	/* close listener sockets */
	close_sock(sock);

	/* finalize signals */
	sig_fin();

	finBidi();
	return;
}

#ifdef WITHOUT_BIDI
/* for test only */
static char *testxmlname = "test.xml";
static FILE *testxmlfp;
#endif

/*
    do bidiStartJob
*/
static int doBidiStartJob(void)
{
    if (initBidi() < 0) return -1;
#ifndef WITHOUT_BIDI
    if ((bidiCanJob) && (job)) {
	/* Only when jobs is started,
	   I call bidiStartJob */
	if ((bidiJob == 0) && (!eprogress)) {
	    if (bidiStartJob(bidiC,job->id/*BIDI_JOB_ID*/) == BIDI_ERROR) {
		procBidiError();
		return -1;
	    }
#ifdef DEBUG
	    fprintf(stderr,"bidiStartJob\n");
#endif
	}
	eprogress = 0;
	bidiJob++;
    }
#endif
    return 0;
}

static int doBidiEndJob(void)
{
#ifndef WITHOUT_BIDI
    if ((bidiCanJob) && (job) && (bidiJob > 0)) {
	int r;

	bidiJob--;
	/* Only when no job is available, I call bidiEndJob */
	if ((bidiJob == 0) && (bidiC != NULL)) {
	    if ((r = bidiEndJob(bidiC)) == BIDI_EPROGRESS) {
		eprogress = 1;
	    } else {
		if (r == BIDI_ERROR) {
		    procBidiError();
		    return -1;
		}
	    }
#ifdef DEBUG
	    fprintf(stderr,"bidiEndJob\n");
#endif
	}
    }
#endif
    return 0;
}

static int doBidiStartWrite(void)
{
#ifndef WITHOUT_BIDI
    if (bidiCanWrite && !bidiInWriting) {
	int r;

	alarm(OPPD_OUTPUT_TIMEOUT);
	r = bidiStartWrite(bidiC);
	alarm(0);
	if (r != BIDI_OK) {
	    procBidiError();
	    return -1;
	}
    #ifdef DEBUG
	fprintf(stderr,"bidiStartWrite\n");
    #endif
	bidiInWriting = 1;
    }
#endif
    return 0;
}

static int doBidiEndWrite(void)
{
#ifndef WITHOUT_BIDI
    if (bidiCanWrite && bidiInWriting) {
	bidiInWriting = 0;
	if (bidiC != NULL) bidiEndWrite(bidiC);
#ifdef DEBUG
	fprintf(stderr,"bidiEndWrite \n");
#endif
    }
#endif
    return 0;
}

int PDBidiStartRead(BidiReadMode idReadMode, char *pLang)
{
    int r = BIDI_OK;

#ifdef DEBUG
    printf("BidiStartRead:idReadMode=%d, pLang=%s\n",idReadMode,pLang);
    fflush(stdout);
#endif
#ifdef WITHOUT_BIDI
    if ((testxmlfp = fopen(testxmlname,"r")) == NULL) {
	fprintf(stderr,"Can't open file %s\n",testxmlname);
	exit(2);
    }
#else
    if (initBidi() < 0) {
	return BIDI_ERROR;
    }
    /* bidiEndWrite if needed */
    if (doBidiEndWrite() < 0) {
	procBidiError();
	return BIDI_ERROR;
    }
    alarm(OPPD_OUTPUT_TIMEOUT);
    r = bidiStartRead(bidiC,idReadMode,pLang);
    alarm(0);
    if (r == BIDI_ERROR || r == BIDI_EINTR) {
	procBidiError();
	return BIDI_ERROR;
    }
#endif

    oppdState |= PDST_ST;

    return r;
}

int PDBidiRead(void *pBuf, int nBufBytes)
{
    int r;

#ifdef DEBUG
    printf("BidiRead:nBufBytes=%d",nBufBytes);
#endif
#ifdef WITHOUT_BIDI
    r = fread(pBuf,1,nBufBytes,testxmlfp);
    if (r < 0) r = 0;
#else
    alarm(OPPD_OUTPUT_TIMEOUT);
    r = bidiRead(bidiC,pBuf,nBufBytes);
    alarm(0);
#endif
#ifdef DEBUG
    printf(",r=%d\n",r);
#endif
    if (r == BIDI_ERROR || r == BIDI_EINTR) {
	procBidiError();
    }
    return r;
}

int PDBidiEndRead(void)
{
#ifdef DEBUG
    printf("BidiEndRead\n");
#endif
#ifdef WITHOUT_BIDI
    fclose(testxmlfp);
#else

    oppdState &= ~PDST_ST;

    if (bidiEndRead(bidiC) == BIDI_ERROR) {
	procBidiError();
	return BIDI_ERROR;
    }
#endif
    return BIDI_OK;
}

int PDBidiCtrl(int idRequest, void *pdata, int nDataBytes)
{
    int r;

#ifdef DEBUG
    printf("bidiCtrl:idRequest=%d, nDataBytes=%d\n",idRequest,nDataBytes);
#endif
#ifdef WITHOUT_BIDI
    printf("************************\n");
    fwrite(pdata,1,nDataBytes,stdout);
    printf("************************\n");
    fflush(stdout);
#else
    if (!bidiCanCtrl) {
	return BIDI_ERROR;
    }
    if (initBidi() < 0) return -1;
    alarm(OPPD_OUTPUT_TIMEOUT);
    r = bidiCtrl(bidiC,idRequest,pdata,nDataBytes);
    alarm(0);
    if (r == BIDI_ERROR) {
	procBidiError();
	return BIDI_ERROR;
    }
#endif
    return BIDI_OK;
}

/*
    set fd_set for select
    sfd: socket file descriptor
*/
static int setFds(int sfd, fd_set *rfds, fd_set *wfds)
{
    int max_fd;

    max_fd = sfd;
    FD_ZERO(rfds);
    FD_ZERO(wfds);
    if (!(oppdState & PDST_ST)) {
	if (want_write) {
	    if (write_fd >= 0) {
		FD_SET(write_fd, wfds);
		if (write_fd > max_fd) max_fd = write_fd;
	    }
	} else if (spl_fd >= 0) {
	    FD_SET(spl_fd, rfds);
	    if (spl_fd > max_fd) max_fd = spl_fd;
	}
    }
    if (sock_fd >= 0) {
	/* PRT_NOST || NOPRT_NOST */
	FD_SET(sock_fd, rfds);
	if (sock_fd > max_fd) max_fd = sock_fd;
    } else {
	FD_SET(sfd, rfds);
    }
    max_fd++; /* add one to change it numbers of fds */
    return max_fd;
}

/*
    process EPROGESS
*/
static int checkEprogress(void)
{
#ifndef WITHOUT_BIDI
    if (eprogress) {
	int r;

	/* try bidiEndJob when EPROGESS mode */
	if ((r = bidiEndJob(bidiC)) != BIDI_EPROGRESS) {
	    eprogress = 0;
	} else {
	    if (r == BIDI_ERROR) {
		procBidiError();
		return -1;
	    }
	}
    }
#endif
    return 0;
}

/*
    check if there is any jobs 
*/
static int checkNewJob(void)
{
    if ((oppdState & PDST_ST) || (sock_fd >= 0)) return 0;

    if ((job = (oppd_job*)oppd_get_job(printer))) {
	struct stat sbuf;

	/* open spool file */
	if ((spl_fd = open(job->spool,
	       O_RDONLY|O_NONBLOCK)) < 0) {
	    return -1;
	}
	if (fstat(spl_fd,&sbuf) < 0) {
	    return -1;
	}
	spl_size = spl_len = sbuf.st_size;
	want_write = 0;
	oppdState |= PDST_PRT;

#ifndef WITHOUT_BIDI
	if (doBidiStartJob() < 0) {
	    return -1;
	}
#endif
    }
    return 0;
}

/*
    read spooled file data into the buffer
*/
static int readSpool(char *buf, int buflen)
{
    if ((read_len = read(spl_fd,buf,buflen)) <= 0) {
	return -1;
    }
    if (doBidiStartWrite() < 0) {
	return -1;
    }
    want_write = 1;
    return 0;
}

/*
    write data in the buffer to the printer
*/
static int writeToPrinter(void)
{
    int wlen;

    if (bidiCanWrite) {
	/* bidiStartWrite if needed */
	if (doBidiStartWrite() < 0) {
	    return -1;
	}
#ifndef WITHOUT_BIDI
	alarm(OPPD_OUTPUT_TIMEOUT);
	wlen = bidiWrite(bidiC,printingBuf,read_len);
	alarm(0);
	if (wlen == BIDI_ERROR || wlen == BIDI_EINTR) {
	    procBidiError();
	    return -1;
	}
#ifdef DEBUG
	fprintf(stderr,"bidiWrite %d\n",wlen);
#endif
#endif
    } else {
	if ((wlen = write(write_fd,printingBuf,read_len)) < 0) {
	    if (errno != EINTR) {
		procBidiError();
	    }
	    return -1;
	}
    }
    spl_len -= wlen;
    if (wlen < read_len) {
	/* remain data to write in buf */
      if (wlen != 0) {
	read_len -= wlen;
	memcpy(printingBuf,printingBuf+wlen,read_len);
      } else {
	usleep(15 * 1000);
      }
    } else {
	/* written all buffered data */
	want_write = 0;
	if (doBidiEndWrite() < 0) {
	    return -1;
	}
	oppdState |= PDST_PRT;
    }
    return 0;
}

/*
    end the printing job 
*/
static void endPrtJob(int delete)
{
    doBidiEndWrite();
    doBidiEndJob();
    if (spl_fd >= 0) {
	close(spl_fd);
	spl_fd = -1;
    }
    want_write = 0;
    if (job) {
	if (delete) oppd_del_job(printer,job);
	job = NULL;
	oppdState &= ~PDST_PRT;
    }
}

/*
 * listen sockets
   main loop of this process
 */
int		listen_sock(
	int		*sock
	)
{
	int ecode = 0;
	fd_set		rfds,wfds;
struct	timeval		tv;
	int		nfds;

	if (listen(*sock, 1) < 0) {
	    return -1;
	}

	while (!force_abort) {
	    int r;
	    struct timeval *tvp;

	    nfds = setFds(*sock,&rfds,&wfds);
	    if (oppdState == PDST_NOPRT_NOST) {
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		tvp = &tv;
	    } else {
		tvp = NULL;
	    }
	    if ((r = select(nfds,&rfds,&wfds,NULL,tvp)) < 0) {
		if (errno == EINTR) {
		    continue;
		}
		/* unrecoverble error */
		ecode = -1;
		endPrtJob(0);
		break;
	    }
	    if (r == 0) { /* state == PDST_NOPRT_NOST */
		/* timeout in select */
		if (checkEprogress() < 0) {
		    endPrtJob(DELJOB_ON_ERR);
		    continue;
		}
		if (!job) { /* check new job */
		    if (checkNewJob() < 0) {
			endPrtJob(DELJOB_ON_ERR);
			continue;
		    }
		}
	    } else {
		if (!(oppdState & PDST_ST)) {
		    if (want_write) {
			if (write_fd >= 0 && FD_ISSET(write_fd, &wfds)) {
			    /* ready for writing to the printer */
			    if (writeToPrinter() < 0) {
				endPrtJob(DELJOB_ON_ERR);
				continue;
			    }
			    if (spl_len <= 0) {
				if (job->copies > 1) {
					job->copies--;
					/* re-start */
					lseek(spl_fd, SEEK_SET, 0);
					spl_len = spl_size;
					want_write = 0;
				} else {
					/* written all spooled data */
					endPrtJob(1);
				}
				continue;
			    }
			}
		    } else if (spl_fd >= 0 && FD_ISSET(spl_fd, &rfds)) {
			/* ready for reading from the spool file */
			if (readSpool(printingBuf,DATA_BUF_LEN) < 0) {
			    endPrtJob(DELJOB_ON_ERR);
			    continue;
			}
		    }
		}
		if (FD_ISSET(*sock,&rfds)) {
		    /* connection request from the status-proc */
		    if ((sock_fd = accept(*sock,NULL,NULL)) < 0) {
			if (errno != EINTR) {
			    /* unrecoverble error */
			    ecode = -1;
			    endPrtJob(DELJOB_ON_ERR);
			    break;
			}
		    }
		    if ((rpcHandle = oprpc_init(sock_fd,sock_fd)) == NULL) {
			close(sock_fd);
			sock_fd = -1;
		    }
		} else if (sock_fd >= 0 && FD_ISSET(sock_fd,&rfds)) {
		    /* request from the status-proc */
		    if (oprpc_doLoop(rpcHandle) < 0) {
			oprpc_destroy(rpcHandle);
			rpcHandle = NULL;
			close(sock_fd);
			sock_fd = -1;
		    }
		}
	    }
	    continue;
	}

	/* clean fds */
	if (sock_fd >= 0) close(sock_fd);
	if (spl_fd >= 0) close(spl_fd);
	/* clean rpc */
	if (rpcHandle != NULL) {
	    oprpc_destroy(rpcHandle);
	}

	return ecode;
}

/*
 * main
 */
int		main(
	int		argc,
	char		**argv
	)
{
	int		ecode = 0;
	int		sock = -1;
	OPPD_ConfEnt	*conf;

	if (argc < 2) exit(-1);
	strcpy(printer, argv[1]);

	/* printer configuration */
	if ((oppd_getConfEntByName(OPPD_CONF_FILE,printer,&conf)) > 0) {
		if (conf->remote) {
		    /* remote printer is not supported */
		    exit(-1);
		}
		strcpy(device, conf->deviceName);
		strcpy(bidi, conf->bidiName);
	} else {
		exit(-1);
	}

	/* initialize */
	if (!ecode) ecode = init(&sock);

 	/* listen sockets */
	if (!ecode) ecode = listen_sock(&sock);

	/* finalize */
	fin(&sock);

	return ecode;
}

