/*

Copyright (c) 2003-2004, AXE, Inc.  All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include "libhttp.h"
#include "oppd-status.h"
#include "opvp_rpc_core.h"
#include "opvp_rpc_client.h"
#include "oppd_conf.h"
#include "oppd_common.h"

#define SOCK_NAME "/tmp/OPPD-%s"

#define MULTI_PSUB

static Conn httpConn;
static char *cPrinter = NULL;
static void *rpcHandle;
static int deviceSock = -1;
static int sockfd = -1;

#ifdef TEST_SOCKET
/*
    create socket for test
*/
static int initSocket(void)
{
    int sock;
    int fd;
    struct sockaddr_in addr;
    int one = 1;

    if ((sock = socket(PF_INET,SOCK_STREAM,0)) < 0) {
	return -1;
    }
    if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&one,sizeof(one)) < 0) {
	return -1;
    }
    bzero(&addr,sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    addr.sin_port = htons(22223);

    if (bind(sock,(struct sockaddr *)&addr,sizeof(addr)) < 0) {
	return -1;
    }
    if (listen(sock,1) < 0) {
	return -1;
    }
    if ((fd = accept(sock,NULL,NULL)) < 0) {
	perror("accept");
	exit(2);
    }
    return fd;
}
#endif

/*
    open output(device) process's socket
*/
static int openDeviceSocket(char *printer)
{
    struct sockaddr_un addr;

#ifndef MULTI_PSUB
    if (cPrinter != NULL) return 0;
#endif
    /* UNIX domain socket */
    if ((deviceSock = socket(PF_UNIX,SOCK_STREAM,0)) < 0) {
	return -1;
    }
    addr.sun_family = AF_UNIX;
    sprintf(addr.sun_path,SOCK_NAME,printer);

    if (connect(deviceSock, (struct sockaddr *)&addr,
          sizeof(addr.sun_family)+strlen(addr.sun_path)) < 0) {
	return -1;
    }
    if ((rpcHandle = oprpc_init(deviceSock,deviceSock)) == NULL) {
	return -1;
    }
    if (cPrinter == NULL) cPrinter = strdup(printer);
    return 0;
}

#ifdef MULTI_PSUB
/*
    close output(device) process's socket
*/
static int closeDeviceSocket(void)
{
    oprpc_destroy(rpcHandle);
    rpcHandle = NULL;
    close(deviceSock);
    return 0;
}
#endif

static void usage(char *cmd)
{
    fprintf(stderr,"Usage:%s -s <socketfd>\n");
    exit(2);
}

static void parseOpt(int argc, char **argv)
{
    int i;

    for (i = 1;i < argc;i++) {
	if (argv[i][0] == '-') {
	    switch (argv[i][1]) {
	    case 's':
		if (++i >= argc) {
		    usage(argv[0]);
		}
		sockfd = atoi(argv[i]);
		break;
	    default:
		usage(argv[0]);
		break;
	    }
	}
    }
    if (sockfd < 0) {
	usage(argv[0]);
    }
}

static int sendNotFound(Conn *conn)
{
    char *msg= "Not found\r\n";
    return HTTPWriteResponse(conn,404,"Not Found",0,NULL,NULL,strlen(msg),msg);
}

static int sendMethodNotAllowed(Conn *conn)
{
    return HTTPWriteResponse(conn,405,"Method Not Allowed",0,NULL,NULL,0,NULL);
}

static int sendInternalError(Conn *conn,int ecode)
{
    char *buf;
    char *names[2],*values[2];
    int i;
    char body[1024];

    sprintf(body,"ecode:%d\r\n",ecode);
    i = 0;
    names[i] = "Content-Type";
    values[i++] = "Text/plain; charset=utf-8";

    return HTTPWriteResponse(conn,500,"Internal Server Error",
       i,names,values,strlen(body),body);
}

static int sendMovedPermanently(Conn *conn,char *uri)
{
    char *names[2],*values[2];
    int i;

    i = 0;
    names[i] = "Location";
    values[i++] = uri;

    return HTTPWriteResponse(conn,301,"Moved Permanently",
       i,names,values,0,NULL);
}

#define BUF_PAGESIZE 4096

/*
    process list command
*/
static int procList(char *p)
{
    int len;
    int buflen;
    int r;
    char *buf;
    char *pbuf;
    OPPD_ConfEnt *ep;
    char *names[1],*values[1];
    int i;

    if (oppd_openConf(OPPD_CONF_FILE) < 0) {
	/* can't open configure file */
	return sendInternalError(&httpConn,OPPD_ERR_OPENCONF);
    }
    if ((buf = malloc(BUF_PAGESIZE*2)) == NULL) {
	return sendInternalError(&httpConn,OPPD_ERR_MEMORY);
    }
    buflen = BUF_PAGESIZE*2;
    pbuf = buf;
    len = 0;
    while ((r = oppd_getConfEnt(&ep)) > 0) {
	int tlen;

	/* calculate required length of space in the buffer */
	tlen = strlen(ep->printerName)+1+strlen(ep->comment)+3;
	if (buflen-len-tlen <= 0) {
	    /* not enough */
	    int add = ((tlen+1)/BUF_PAGESIZE)*BUF_PAGESIZE;

	    /* extend buffer */
	    if ((buf = realloc(buf,buflen+add)) == NULL) {
		return sendInternalError(&httpConn,OPPD_ERR_MEMORY);
	    }
	    buflen += add;
	    pbuf =  buf+len;
	}
	sprintf(pbuf,"%s %s\r\n",ep->printerName,ep->comment);
	pbuf += tlen-1;
	len += tlen-1;
    }
    oppd_endConfEnt();
    if (r < 0) {
	free(buf);
	return sendInternalError(&httpConn,OPPD_ERR_READCONF);
    }
    i = 0;
    names[i] = "Content-Type";
    values[i++] = "Text/plain; charset=utf-8";
    r = HTTPWriteResponse(&httpConn,200,"OK",i,names,values,len,buf);
    free(buf);
    return r;
}

/*
    check printer name
    return
    -1: internal error
    0: error or remote printer
    1: local printer
*/
static int checkPrinterName(char *printer)
{
    OPPD_ConfEnt *ep;
    int r;

    if ((r = oppd_getConfEntByName(OPPD_CONF_FILE,printer,&ep)) <= 0) {
	if (r == 0) {
	    return sendNotFound(&httpConn);
	}
	return sendInternalError(&httpConn,OPPD_ERR_READCONF);
    }
    if (ep->remote) {
	return sendMovedPermanently(&httpConn,ep->deviceName);
    }
    return 1;
}

/*
    process status command
    p: printer name string
*/
static int procGetStatus(char *p)
{
    int idReadMode = 1;
    char *pLang = NULL;
    char *printer;
    char *r;
    int rval;
    char *buf;
    char *pbuf;
    char *names[1],*values[1];
    int i = 0;
    int fd;
    int len;
    int buflen;
    int rv;

    if (p == NULL) {
	/* no printer name */
	return sendNotFound(&httpConn);
    }
    printer = p;
    /* get options */
    if ((r = strchr(p,'?')) != NULL) {
	/* there are some options */
	char *name, *value;

	*r++ = '\0';
	while ((name = HTTPParseOpt(r,&value,&r)) != NULL) {
	    if (strcmp(name,"idReadMode") == 0) {
		idReadMode = atoi(value);
	    } else if (strcmp(name,"LANG") == 0) {
		pLang = value;
	    }
	}
    }
    if ((rv = checkPrinterName(printer)) <= 0) {
	return rv;
    }
    if (cPrinter != NULL && strcmp(printer,cPrinter) != 0) {
	/* if already open device socket and printer is different,
	  it should not be happened */
	return sendNotFound(&httpConn);
    }
    /* if device scoket is not open, I open it */
    if (openDeviceSocket(printer) < 0) {
	return sendNotFound(&httpConn);
    }
    /* call output process's bidiStartRead */
    if ((rval = CStubBidiStartRead(rpcHandle,idReadMode,pLang)) < 0) {
#ifdef MULTI_PSUB
	closeDeviceSocket();
#endif
	return sendInternalError(&httpConn,rval);
    }
    if ((buf = malloc(BUF_PAGESIZE*2)) == NULL) {
#ifdef MULTI_PSUB
	closeDeviceSocket();
#endif
	return sendInternalError(&httpConn,OPPD_ERR_MEMORY);
    }
    buflen = BUF_PAGESIZE*2;
    pbuf = buf;
    len = 0;
    /* call output process's bidiRead  */
    while ((rval = CStubBidiRead(rpcHandle,pbuf,BUF_PAGESIZE)) > 0) {
	pbuf += rval;
	len += rval;
	if (buflen-len < BUF_PAGESIZE) {
	    /* extend buffer */
	    if ((buf = realloc(buf,buflen+BUF_PAGESIZE)) == NULL) {
#ifdef MULTI_PSUB
		closeDeviceSocket();
#endif
		return sendInternalError(&httpConn,OPPD_ERR_MEMORY);
	    }
	    buflen += BUF_PAGESIZE;
	    pbuf =  buf+len;
	}
    }
    /* call output process's bidiEndRead  */
    rval = CStubBidiEndRead(rpcHandle);
#ifdef MULTI_PSUB
    closeDeviceSocket();
#endif
    if (rval < 0) {
	free(buf);
	return sendInternalError(&httpConn,rval);
    }
    i = 0;
    names[i] = "Content-Type";
    values[i++] = "Text/xml; charset=utf-8";
    rval = HTTPWriteResponse(&httpConn,200,"OK",i,names,values,len,buf);
    free(buf);
    return rval;
}

/*
    process ctrl command
    p: printer name string
*/
static int procCtrl(char *p)
{
    int idRequest = 0;
    char *printer;
    char *r;
    int rval;
    char *name, *value;
    int rv;

    if (p == NULL) {
	/* no printer name */
	return sendNotFound(&httpConn);
    }
    printer = p;
    if ((r = strchr(p,'?')) == NULL) {
	/* there is no option. this is error,
	   because idRequest option is mandatory  */
	return sendNotFound(&httpConn);
    }

    *r++ = '\0';
    while ((name = HTTPParseOpt(r,&value,&r)) != NULL) {
	if (strcmp(name,"idRequest") == 0) {
	    idRequest = atoi(value);
	}
    }
    if ((rv = checkPrinterName(printer)) <= 0) {
	return rv;
    }
    if (cPrinter != NULL && strcmp(cPrinter,printer) != 0) {
	/* if already open device socket and printer is different,
	  it should not be happened */
	return sendNotFound(&httpConn);
    }
    /* if device scoket is not open, I open it */
    if (openDeviceSocket(printer) < 0) {
	return sendNotFound(&httpConn);
    }

    /* call output process's bidiCtrl  */
    rval = CStubBidiCtrl(rpcHandle, idRequest,httpConn.body,
          httpConn.length);
#ifdef MULTI_PSUB
    closeDeviceSocket();
#endif
    if (rval < 0) {
	return sendInternalError(&httpConn,rval);
    }

    return HTTPWriteResponse(&httpConn,200,"OK",0,NULL,NULL,0,NULL);
}

/*
    process one request from PSUB
*/
static int procOneReq(void)
{
    char *uri = httpConn.uri;
    char *p;

    if (uri == NULL) return -1;
    if (uri[0] != '/') {
	sendNotFound(&httpConn);
	return 0;
    }
    /* split a command string and a printer name */
    if ((p = strchr(uri+1,'/')) != NULL) {
	*p++ = '\0';
    }
    /* now p point to the printer name */
    if (strcasecmp("get",httpConn.method) == 0) {
	if (strcmp("list",uri+1) == 0) {
	    return procList(p);
	} else if (strcmp("status",uri+1) == 0) {
	    return procGetStatus(p);
	} else {
	    sendNotFound(&httpConn);
	    return 0;
	}
    } else if (strcasecmp("post",httpConn.method) == 0) {
	if (strcmp("status",uri+1) == 0) {
	    return procCtrl(p);
	} else {
	    sendNotFound(&httpConn);
	    return 0;
	}
    } else {
	sendMethodNotAllowed(&httpConn);
	return 0;
    }
    return 0;
}

int main(int argc, char **argv)
{
#ifdef TEST_SOCKET
    if ((sockfd = initSocket()) < 0) {
	perror("initSocket");
	exit(2);
    }
#else
    parseOpt(argc,argv);
#endif
    if (HTTPInit(sockfd,&httpConn) < 0) {
	fprintf(stderr,"HTTPInit error\n");
	exit(2);
    }
    while (HTTPGetReq(&httpConn) >= 0) {
	if (procOneReq() < 0) {
	    break;
	}
    }
    HTTPDestroy(&httpConn);
    close(sockfd);
    return 0;
}
